;;;*************************************************************************
;; ---------------------
;; UNIFICATION PROCEDURE
;; ---------------------
(module unification
	(include "setheo.sch")
	(export (unbind-vars) (ref term) (unify-lists k l) (unify-terms s t) (extension-unify-lists k l) 
		(unbind-collect-vars variables))
	(import (globals "globals.scm")))

(define (ref term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (non-variable term) (free-variable term))
     term
     (ref (car term))))

(define (unify-lists k l)
;; unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     #t
     (if (unify-terms (ref (car k)) (ref (car l)))
	 (unify-lists (cdr k) (cdr l))
	 #f)))

(define (unify-terms s t)          
;; unifies terms s and t. 
;; If unification succeeds, #t is returned, else #f.
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (unbind s t)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
	         (unbind t s)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (unify-lists (cdr s) (cdr t))
                 (unbind s t))))))

(define (occurs-list var termlist)              
;; returns #t if var occurs in termlist, and #f otherwise
 (if (null? termlist)
     #f
     (if (occurs-term var (ref (car termlist)))
         #t
         (occurs-list var (cdr termlist)))))

(define (occurs-term var term)       
;; returns #t if var occurs in term, and #f otherwise
 (if (eq? var term) 
     #t
     (and (non-variable term) (occurs-list var (cdr term)))))

;;;END *********************************************************************

;;;*************************************************************************
;; -------------------------------
;; OPTIMIZED UNIFICATION PROCEDURE
;; -------------------------------

;; For extension steps it is known that initially the two termlists are
;; variable disjoint. Under these conditions the unification can be 
;; optimized such that certain occurs-checks can be omitted, as follows.
;; We assume that for bindings always the first argument position (the k 
;; or s-position, i.e. the entry literal) is preferred; also, we suppose 
;; that no ref-cells are in the terms at the first argument position. 
;; 
;; Initially, the fully optimized procedures extension-unify-* are called.
;; As long as no binding is performed, we remain in this routines.
;;     THE REST IS TO BE CORRECTED ...
;; If the first binding is performed, say, for variable x, then
;; 1. if x is on the second argument position (l or t-position, i.e. from 
;;    the calling subgoal), then we move to standard unification for the 
;;    rest of the unification;
;; 2. otherwise (x must be a free variable) we move to the semi-optimized
;;    procedures semi-extension-* which work as follows:
;;    only free variables at the first argument positions may be bound 
;;    without performing occurs-checks; whenever a variable at the second 
;;    argument position is bound, we move to standard unification.

(define (extension-unify-lists k l)
;; like unify-lists, but fully optimized for extension steps;
;; it is assumed that no binding has been performed yet
 (if 
  (null? k)
  #t
  (if (extension-unify-terms (car k) (ref (car l)))
      (cond ((eq? optimized-unification 0) 
             (extension-unify-lists (cdr k) (cdr l)))
            ((eq? optimized-unification 1) 
             (semi-extension-unify-lists (cdr k) (cdr l)))
	    (else (unify-lists (cdr k) (cdr l))))
      #f)))

(define (extension-unify-terms s t)          
;; like unify-terms, but fully optimized for extension steps;
;; it is assumed that no binding has been performed yet
 (if (variable s)
     (begin (set! optimized-unification 1) (bind s t))
     (if (variable t)
         (begin (set! optimized-unification #f) (bind t s))
         (if (eq? (car s) (car t))
	     (extension-unify-lists (cdr s) (cdr t))
             (unbind s t)))))

(define (semi-extension-unify-lists k l)
;; like unify-lists, but semi-optimized for extension steps;
;; it is assumed that only bindings for variables at the first argument
;; positions have been performed yet
 (cond ((null? k) #t)
       ((free-variable (car k)) (bind (car k) (ref (car l)))
        (semi-extension-unify-lists (cdr k) (cdr l)))
       ((variable (car k)) 
        (if (unify-terms (ref (car k)) (ref (car l)))
            (semi-extension-unify-lists (cdr k) (cdr l))
            #f))
       ((semi-extension-unify-terms (car k) (ref (car l)))
        (if (eq? 1 optimized-unification)
            (semi-extension-unify-lists (cdr k) (cdr l))
            (unify-lists (cdr k) (cdr l))))
       (else #f)))

(define (semi-extension-unify-terms s t)          
;; like unify-terms, but semi-optimized for extension steps;
;; it is assumed that only bindings for variables at the first argument
;; positions have been performed yet
 (if (variable t)
     (if (occurs-list t (cdr s))
	 (unbind t s)
         (begin (set! optimized-unification #f) (bind t s)))
     (if (eq? (car s) (car t))
	 (semi-extension-unify-lists (cdr s) (cdr t))
         (unbind s t))))

;;;END**********************************************************************
;;;*************************************************************************
;; ---------------------------------
;; DIFFERENCE UNIFICATION PROCEDURES
;; ---------------------------------
;; This unification procedure always succeeds. If two terms are unifiable,
;; the procedure behaves like ordinary unification. If, however, a failure
;; point in the standard unification would be reached, this procedure stores 
;; the non-unifiable pair of terms and proceeds.

(define (difference-unify-lists k l)
;; difference-unifies termlists k and l.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (pair? k)
     (begin (difference-unify-terms (ref (car k)) (ref (car l)))
	    (difference-unify-lists (cdr k) (cdr l)))))

(define (difference-unify-terms s t)          
;; difference-unifies terms s and t. 
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (store-difference s t)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
                 (store-difference t s)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (difference-unify-lists (cdr s) (cdr t))
                 (store-difference s t))))))

;;;END**********************************************************************
;;;*************************************************************************
;; -----------------------------
;; TRAIL MODIFICATION PROCEDURES

(define (unbind-vars)
;; makes free all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) (trail-previous) (unbind-vars))
     (begin (trail-previous) #f)))

(define (unbind-collect-vars variables)
;; makes free and collects all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
	    (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (unbind-collect-vars variables))
     (begin (trail-previous) variables)))

(define (last-binding-variables)
 (let ((old-trail trail) variables)
      (set! variables (last-binding-vars '()))
      (set! trail old-trail)
      variables))

(define (last-binding-vars variables)
 (if (trail-content)
     (begin (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (last-binding-vars variables))
     variables))