;;;*************************************************************************
;;;module name: prover.scm
;;;author     : Reinhold Letz, TU Munich
;;;date       : 29.10.98
;;;description: propositional decision procedure using the semantic tree 
;;;		calculus (DPLL-kernel)
;;;remarks    : input is a sequence of clauses;
;;;             when the flag -geninput is set,
;;;             1) an input clause must be a list of the form 
;;;                ( positive_literals negative_literals ) 
;;;                where positive_literals and positive_literals are lists
;;;                of arbitrary scheme objects
;;;             2) otherwise, a clause is a list of integers different from 0,
;;;                negation is expressed with arithmetic negation
;;;*************************************************************************

(define (semprop args)
 (input-formula args)
 (set! output (evaluate 0 0 0))
 (if (> inference-limit 0) (random-evaluate))
 (if (= nr-models 0)
     (begin (display "unsatisfiable: ") (display output))
     (begin (display "satisfiable: ") 
	    (if (and allmodels
		     (or (= verbose 0) (and (= verbose 1) (> nr-models 1))))
		(begin (display "nr of models: ") (display nr-models)
		       (display ", ")))
	    (display end-inference)))
 (display " branches")
 (if (> attempts 1) (begin (display " in attempt ") (display attempts)))
 (newline)
)

(define (random-evaluate)
 (if (and (not output) (= 0 nr-models))
     (begin (set! random-branch (- random-branch 1))
            (set! attempts (+ attempts 1))
;            (display random-branch) (newline)
            (set! inference-limit (+ inference-limit
                (quotient (* inference-limit-increase inference-limit) 100)))
            (set! unit-index -1) (insert-units-in-vector unit-list)
            (set! output (evaluate 0 0 0))
            (random-evaluate))))

(define (input-formula args)
; (display "Argument list: ") (display args) (newline)
 (update-defaults args)
 (set! clauses-vector (make-vector (+ max-cl-length 1) '()))
 (set! clauses-weight-vector (make-vector (+ max-cl-length 1)))
 (set! infile (open-input-file (car (last args))))
 (if split 
     (set! input-size (+ hash-size max-def-number))
     (set! input-size hash-size))
; (display "input-size: ") (display input-size) (newline)
 (if general-input
     (begin (set! hash-table (make-vector hash-size #f))
            (load-formula-general infile))
     (load-formula infile))
 (compile)
)

(define (load-formula port)
 (set! interm (read port))
 (if (eof-object? interm)
     (close-input-port infile)
     (begin (set! clause-nr (+ 1 clause-nr))
	    (clause-insert)
            (load-formula port))))

(define (load-formula-general port)
 (set! interm (read port))
 (if (eof-object? interm)
     (close-input-port infile)
     (begin (set! interm (hashify-clause interm))
	    (set! clause-nr (+ 1 clause-nr))
            (clause-insert)
            (load-formula-general port))))

(define (evaluate atom-index inference branch-depth)
;d; (newline) (display "INFERENCE ") (display inference) (display " FALSIFY-LEFT: ") 
;d; (newline) (display "UI 1: ") (display unit-index) 
(if 
 (and (> inference-limit 0) (> inference inference-limit))
 (begin (display inference) (display " ") #f) 
 (let (atom sign clause clauses true-clauses
       new-inference new-inference2 (local-used-inference inference))
  (if (>= unit-index 0) ; a unit clause exists
      (begin (set! inference (+ 1 inference))
        (set! clause (vector-ref unit-vector unit-index))
        (set! atom (get-literal (c-pos-literals clause)))
        (if atom 
            (set! sign 1) 
            (begin (set! atom (get-literal (c-neg-literals clause)))
                   (set! sign 0))) 
        (rplac-atom-inference (c-pos-literals clause) inference)
        (rplac-atom-inference (c-neg-literals clause) inference)
;d; (display " UNIT: ") (display-clause clause) 
        (set! unit-index (- unit-index 1))
;d; (newline) (display "INFERENCE ") (display inference)  
;d; (display " LEFT-UNIT-RIGHT-EVALUATE: ") 
	(if (= sign 0) ;; negative valuation of atom to be checked
	    (begin (set! clauses (a-pos-clauses atom))
;d;                    (display "-") 
;                    (set! literal-index (- atom-index))
		   (set-a-sign atom 0)
		   (set! true-clauses (a-neg-clauses atom)))
	    (begin (set! clauses (a-neg-clauses atom))
;                    (set! literal-index atom-index)
		   (set-a-sign atom 1)
		   (set! true-clauses (a-pos-clauses atom))))
;d; (display (a-name atom)) 
;d; (newline) (display "UI 2: ") (display unit-index) 
	(set! stop-clause #f) 
;	 (set! current-atom atom) 
	(decrease-counters clauses)
;d; (newline) (display "UI 3: ") (display unit-index) 
	(if stop-clause 
	    (begin      
;d; (display " STOP CLAUSE: ") (display-clause stop-clause) 
	      (increase-counters clauses) ;; do not exchange with next line!
	      (set-a-sign atom #f) ; atom is made available
	      (set! inference (+ 1 inference))
;d; (newline) (display "U1 4: ") (display unit-index) 
	      (rplac-atom-inference (c-pos-literals stop-clause) inference)
	      (rplac-atom-inference (c-neg-literals stop-clause) inference)
	      (set! glob-used-inference local-used-inference)
	      inference)
	    (begin  
;d; (display " NO STOP CLAUSE") 
	      (disactivate true-clauses atom)
;d; (newline) (display-all-clauses clause-list) 
	      (find-active) ; moves unit-index to next unit clause or -1
	      (set! new-inference (evaluate atom-index inference branch-depth))
;	       (set! current-atom atom) 
	      (reactivate true-clauses atom) 
	      (incr-counters clauses) ; clause lengths are incremented
	      (set-a-sign atom #f) ;; do not perform before previous 2 lines
	      (if (>= (a-last-inference atom) glob-used-inference)
		  (set! glob-used-inference local-used-inference))
	      new-inference)))
      (begin (set! atom-index (next-atom-index atom-index))
;d; (display " NO UNIT") 
       (if atom-index ; no unit clause, but there are still atoms 
	   (branching atom-index inference atom sign clauses true-clauses
  	      new-inference new-inference2 local-used-inference branch-depth)
	   (begin (set! glob-used-inference local-used-inference)
		  (set! inference (+ 1 inference))
		  (model inference))))))))

(define (find-active)
 (if (>= unit-index 0)
     (if (c-activating-atom (vector-ref unit-vector unit-index))
         (begin (set! unit-index (- unit-index 1)) (find-active)))))

(define (decrease-counters clauses)
;; decreases counters with considering stop clause
;; modifies the vector of units resulting from the specialization if no
;; stop clause exists; if a stop clause exists, stop-clause is set
 (if (pair? clauses) 
     (cond ((c-activating-atom (car clauses)) ; clause is not active
            (decrease-counters (cdr clauses)))
           ((= 1 (c-length (car clauses))) ; complement falsifies a unit clause
            (set! stop-clause (car clauses)))
           ((= 2 (c-length (car clauses))) ; a new unit clause is produced
            (set-c-length (car clauses) 1)
;            (set-c-unit-atom-index (car clauses) literal-index)
            (set! unit-index (+ unit-index 1))
;d; (newline) (display "DECR: ") (display unit-index) 
;            (if (= unit-index unit-vector-length) (increase-unit-vector))
            (vector-set! unit-vector unit-index (car clauses)) 
            (decrease-counters (cdr clauses)))
           (else (set-c-length (car clauses) (- (c-length (car clauses)) 1))
                 (dyn-incr-weight1 (car clauses))
		 (decrease-counters (cdr clauses))))))

(define (decr-counters clauses)
;; decreases counters without considering stop clause
;; modifies the vector of units resulting from the specialization 
 (if (pair? clauses) 
     (begin
       (if (not (c-activating-atom (car clauses))) ;; clause is active
	   (begin
	     (if (= 2 (c-length (car clauses))) ;; a new unit clause produced
		 (begin (set-c-length (car clauses) 1)
;                     (set-c-unit-atom-index (car clauses) literal-index)
			(set! unit-index (+ unit-index 1))
;d; (newline) (display "DEC: ") (display unit-index) 
			(if (= unit-index unit-vector-length)
			    (increase-unit-vector))
			(vector-set! unit-vector unit-index (car clauses)))
		 (set-c-length (car clauses) (- (c-length (car clauses)) 1)))
	     (dyn-incr-weight1 (car clauses))))
       (decr-counters (cdr clauses)))))

(define (increase-freq-pos literals)
 (if (pair? literals)
     (begin (if (not (a-sign (car literals)))
		(set-a-pos-frequency (car literals)
				 (+ weight (a-pos-frequency (car literals)))))
	    (increase-freq-pos (cdr literals)))))
 
(define (increase-freq-neg literals)
 (if (pair? literals)
     (begin (if (not (a-sign (car literals)))
		(set-a-neg-frequency (car literals)
				 (+ weight (a-neg-frequency (car literals)))))
	    (increase-freq-neg (cdr literals)))))
 
(define (increase-counters clauses)
;; increases counters with considering stop clause
 (if (pair? clauses) 
     (if (not (eq? stop-clause (car clauses)))
         (if (c-activating-atom (car clauses)) ;; clause is not active
             (increase-counters (cdr clauses))
             (begin (set-c-length (car clauses) (+ 1 (c-length (car clauses))))
		    (dyn-decr-weight1 (car clauses))
		    (increase-counters (cdr clauses)))))))

(define (incr-counters clauses)
;; increases counters without considering stop clause
 (if (pair? clauses) 
     (begin 
       (if (not (c-activating-atom (car clauses))) ;; clause is active
	   (begin (set-c-length (car clauses) (+ 1 (c-length (car clauses))))
		  (dyn-decr-weight1 (car clauses))))
       (incr-counters (cdr clauses)))))

(define (increase-unit-vector)
 (set! old-unit-vector unit-vector)
 (set! unit-vector (make-vector (* 2 unit-vector-length)))
 (copy-unit-vector 0)
 (set! unit-vector-length (* 2 unit-vector-length)))

(define (copy-unit-vector index)
 (if (< index unit-vector-length)
     (begin (vector-set! unit-vector index (vector-ref old-unit-vector index))
            (copy-unit-vector (+ 1 index)))))

(define (disactivate clauses atom)
 (if (pair? clauses) 
     (begin (if (not (c-activating-atom (car clauses))) ;; clause is active
                (set-c-activating-atom (car clauses) atom))
            (dyn-decr-weight (car clauses))
	    (disactivate (cdr clauses) atom))))

(define (reactivate clauses atom)
 (if (pair? clauses) 
     (begin (if (eq? (c-activating-atom (car clauses)) atom)
                (set-c-activating-atom (car clauses) #f)) ;; clause reactivated
            (dyn-incr-weight (car clauses))
            (reactivate (cdr clauses) atom))))

(define (rplac-atom-inference atoms inference)
 (if (pair? atoms) 
     (begin (set-a-last-inference (car atoms) inference)
            (rplac-atom-inference (cdr atoms) inference))))

(define (get-literal clause-half)
 (if (pair? clause-half)
     (if (a-sign (car clause-half))
         (get-literal (cdr clause-half))
         (car clause-half))
     #f))

(define (next-atom-index atom-index)
 (if (< atom-index nr-atoms) 
     (if (a-sign (vector-ref atom-vector atom-index))
         (next-atom-index (+ 1 atom-index))
         atom-index)
     #f))

(define (select-atom-dyn frequency other-index steps)
 (if 
  (> steps 0)
  (begin (set! other-index (next-atom-index other-index))
;(display "V")
   (if other-index
       (begin (set! other-atom (vector-ref atom-vector other-index))
	(set! other-frequency 
	      (if max-frequency
		  (max (a-pos-frequency other-atom)
		       (a-neg-frequency other-atom))
		  (+ (a-pos-frequency other-atom)
		     (a-neg-frequency other-atom))))
	(if (< frequency other-frequency)
            (begin (set! selected-index other-index)
;(display "-")
	      (select-atom-dyn other-frequency (+ other-index 1) (+ steps 1)))
	    (select-atom-dyn frequency (+ other-index 1) (+ steps 1)))
)))))

(define (select-atom-dynamic index frequency nr-t-clauses other-index steps)
 (if 
  (> steps 0) 
  (begin (set! other-index (next-atom-index other-index))
   (if other-index
       (begin (set! other-atom (vector-ref atom-vector other-index))
        (if (= frequency (a-frequency other-atom))
            (begin (set! nr-t-clauses-next 
                       (+ (nr-true-clauses (a-pos-clauses other-atom) 0)
                          (nr-true-clauses (a-neg-clauses other-atom) 0)))
             (if (< nr-t-clauses-next nr-t-clauses)
                 (begin 
;(display "Switching ") (display (a-name (vector-ref atom-vector index)))
;(display " and ") (display (a-name other-atom))
;(newline) 
                        (vector-set! atom-vector other-index 
                                     (vector-ref atom-vector index))
                        (vector-set! atom-vector index other-atom)
                        (select-atom-dynamic index frequency nr-t-clauses-next
                                            (+ other-index 1) (+ steps 1)))
                 (select-atom-dynamic index frequency nr-t-clauses 
                                            (+ other-index 1) (+ steps 1))))
))))))

(define (nr-true-clauses clauses sum)
 (if (pair? clauses)
     (begin (if (c-activating-atom (car clauses)) (set! sum (+ sum 1)))
            (nr-true-clauses (cdr clauses) sum))
     sum))

(define (model inference)
 (set! end-inference inference)
 (set! nr-models (+ 1 nr-models))
 (if (or (= verbose 2) (and (= 1 nr-models) (= verbose 1)))
     (begin (if (not also-neg-in-model) 
		(display "positive part of "))
	    (display "model ")
	    (display nr-models)
	    (display ":")
	    (newline)
	    (print-model line atom-list (- line 1))
	    (newline)))
 (if allmodels 
     (begin (rplac-atom-inference atom-list inference) inference) 
     #f))

(define (print-model nr atom-list line)
 (if (pair? atom-list)
     (begin 
       (if (= 1 (a-sign (car atom-list)))
           (begin (if (= 0 nr) 
                      (begin (newline) (set! nr line))
                      (set! nr (- nr 1)))
                  (display " ") (display (a-realname (car atom-list))))
           (if also-neg-in-model 
               (begin (if (= 0 nr) 
                          (begin (newline) (set! nr line))
                          (set! nr (- nr 1)))
                      (display " -") (display (a-realname (car atom-list))))))
       (print-model nr (cdr atom-list) line))))

(define (last l)
  (if (null? (cdr l))
      l
      (last (cdr l))))
