;;;*************************************************************************
;;;module name: semprop.sch
;;;author     : Reinhold Letz, TU Munich
;;;date       : 29.10.98
;;;description: propositional decision procedure using the semantic tree 
;;;		calculus (DPLL-kernel)
;;;remarks    : literals are integers and must not occur multiply in a clause.
;;;compiling  : yes
;;;*************************************************************************

;; -------------------------------------------------------------------------
;; internal data structures:
;; 
;; clause: vector of four elements
;; 1st element: atom which activates the clause [#f or atom]
;; 2nd element: current clause length [positive integer]
;; 3rd element: list of positive atoms
;; 4th element: list of negative atoms
;; 5th element: clause number
;;
;; atom: vector of seven elements
;; 1st element: atom is currently unevaluated [#f] or evaluated [0 or 1] 
;; 2nd element: clauses with positive occurrence of atom
;; 3rd element: clauses with negative occurrence of atom
;; 4th element: last inference using assignment of atom
;; 5th element: positive atom frequency in formula
;; 6th element: negative atom frequency in formula
;; 7th element: atom name
;; 8th element: atom sign
;; 
;; -------------------------------------------------------------------------

;; macros for accessing and modifying basic data structures
;;
(define-macro (c-activating-atom clause) `(vector-ref ,clause 0))
(define-macro (c-length clause)          `(vector-ref ,clause 1))
(define-macro (c-pos-literals clause)    `(vector-ref ,clause 2))
(define-macro (c-neg-literals clause)    `(vector-ref ,clause 3))
(define-macro (c-number clause)          `(vector-ref ,clause 4))
(define-macro (c-unit-atom-index clause) `(vector-ref ,clause 5))

(define-macro (set-c-activating-atom clause val) `(vector-set! ,clause 0 ,val))
(define-macro (set-c-length clause val)          `(vector-set! ,clause 1 ,val))
(define-macro (set-c-pos-literals clause val)    `(vector-set! ,clause 2 ,val))
(define-macro (set-c-neg-literals clause val)    `(vector-set! ,clause 3 ,val))
(define-macro (set-c-number clause val)          `(vector-set! ,clause 4 ,val))
(define-macro (set-c-unit-atom-index clause val) `(vector-set! ,clause 5 ,val))

(define-macro (a-pos-clauses atom)       `(vector-ref ,atom 0))
(define-macro (a-neg-clauses atom)       `(vector-ref ,atom 1))
(define-macro (a-last-inference	atom)    `(vector-ref ,atom 2))
(define-macro (a-pos-frequency atom)     `(vector-ref ,atom 3))
(define-macro (a-neg-frequency atom)     `(vector-ref ,atom 4))
(define-macro (a-frequency atom)         `(vector-ref ,atom 5))
(define-macro (a-name atom)              `(vector-ref ,atom 6))
(define-macro (a-sign atom)              `(vector-ref ,atom 7))

(define-macro (a-realname atom)
`(if general-input
     (vector-ref hash-table (a-name ,atom))
     (a-name ,atom)))

(define-macro (set-a-pos-clauses atom val)    `(vector-set! ,atom 0 ,val))
(define-macro (set-a-neg-clauses atom val)    `(vector-set! ,atom 1 ,val))
(define-macro (set-a-last-inference atom val) `(vector-set! ,atom 2 ,val))
(define-macro (set-a-pos-frequency atom val)  `(vector-set! ,atom 3 ,val))
(define-macro (set-a-neg-frequency atom val)  `(vector-set! ,atom 4 ,val))
(define-macro (set-a-frequency atom val)      `(vector-set! ,atom 5 ,val))
(define-macro (set-a-name atom val)           `(vector-set! ,atom 6 ,val))
(define-macro (set-a-sign atom val)           `(vector-set! ,atom 7 ,val))

(define-macro (make-atom atom-name) `(vector '() '() 0 0 0 0 ,atom-name #f))

(define-macro (branching atom-index inference atom sign clauses true-clauses
	       new-inference new-inference2 local-used-inference branch-depth)
; a proper semantic tree branching is performed
`(begin 
;d; (newline) (display "INFERENCE ") (display inference) (display " BRANCHING: ") 
  (if (<= max-branch-depth ,branch-depth)
      (begin (set! dynamic #f)
	     (set! ,atom (vector-ref atom-vector ,atom-index)))
      (begin (set! selected-index ,atom-index)
	(select-atom ,atom-index)
	(if (not (= ,atom-index selected-index))
	    (begin (set! ,atom (vector-ref atom-vector selected-index)))
	    (set! ,atom (vector-ref atom-vector ,atom-index)))))
  (if (select-sign ,atom ,inference) ; returns #t if positive valuation first
      (begin (set! ,sign 1) (set! ,clauses (a-neg-clauses ,atom))
;             (set! literal-index ,atom-index)
             (set-a-sign ,atom 1)
             (set! ,true-clauses (a-pos-clauses ,atom)))
      (begin (set! ,sign 0) (set! ,clauses (a-pos-clauses ,atom))
;             (set! literal-index (- ,atom-index))
             (set-a-sign ,atom 0)
;d; (display "-") 
             (set! ,true-clauses (a-neg-clauses ,atom))))
;d; (display (a-name ,atom)) 
;d;(newline) (display unit-index) 
;  (set! current-atom ,atom) 
  (decr-counters ,clauses)
  (disactivate ,true-clauses ,atom)
;d; (newline) (display-all-clauses clause-list) 
  (set! glob-used-inference ,inference)
  (find-active) ; moves unit-index to next unit clause or -1
  (set! ,new-inference (evaluate ,atom-index ,inference (+ 1 ,branch-depth)))
;  (set! current-atom ,atom) 
  (reactivate ,true-clauses ,atom) ; clauses are reactivated
  (incr-counters ,clauses) ; clause lengths are incremented
  (set-a-sign ,atom #f) ; atom made available: always after previous 2 lines!
  (if ,new-inference ; no model was found in the left subtree
      (if (>= (a-last-inference ,atom) glob-used-inference); literal used
          (begin (set! unit-index -1)
;          (set! literal-index (- ,atom-index))
		 (if (= ,sign 1) ; set sign always before decr. and disact.
		     (begin (set! ,sign 0) (set-a-sign ,atom 0))
		     (begin (set! ,sign 1) (set-a-sign ,atom 1)))
;		 (set! current-atom ,atom) 
		 (decr-counters ,true-clauses) 
		 (disactivate ,clauses ,atom)
		 (find-active) ; moves unit-index to next unit clause or -1
		 (set! ,local-used-inference glob-used-inference)
		 (set! ,new-inference2 (evaluate ,atom-index ,new-inference
						 (+ 1 ,branch-depth)))
;		 (set! current-atom ,atom) 
		 (reactivate ,clauses ,atom) ; clauses are reactivated
		 (incr-counters ,true-clauses) ; clause lengths are incremented
		 (set-a-sign ,atom #f) ;; do not exchange with previous 2 lines
		 (if (>= (a-last-inference ,atom) glob-used-inference);lt. used
		     (set! glob-used-inference ,local-used-inference))
		 (if (= ,branch-depth max-branch-depth) (set! dynamic #t))
		 ,new-inference2)
	  (begin (if (= ,branch-depth max-branch-depth) (set! dynamic #t))
		 ,new-inference))
      (begin (if (= ,branch-depth max-branch-depth) (set! dynamic #t))
	     #f))))

(define-macro (select-atom atom-index) 
`(if dynamic                ;(> dynamic-atom-selection 0)
     (begin (set! atom0 (vector-ref atom-vector selected-index))
       (select-atom-dyn (if max-frequency
			    (max (a-pos-frequency atom0)
				 (a-neg-frequency atom0))
			    (max greatest-int
				 (+ (a-pos-frequency atom0)
				    (a-neg-frequency atom0))))
			; (+ (nr-true-clauses (a-pos-clauses atom) 0)
                        ; (nr-true-clauses (a-neg-clauses atom) 0))
			(+ selected-index 1)
                        dynamic-atom-selection))))

(define-macro (select-sign atom inference)
`(if (and (> random-branch 0) 
          (= (quotient (a-pos-frequency ,atom) freq-divisor)
             (quotient (a-neg-frequency ,atom) freq-divisor))
          (> random-percentage
             (modulo (modulo random-branch (+ ,inference 1)) 100)))
     (if (= 0 (modulo (modulo random-branch (+ ,inference 1)) 2))
         #f
         #t)
     (if neg-first
	 #f
	 (if pos-first
	     #t
	     (if (<= (a-pos-frequency ,atom) (a-neg-frequency ,atom))
		 #f
		 #t)))))
 
;(define-macro (clause-insert)
;`(begin (set! tautology #f) (set! cl-length 1) (prune-clause)
;   (if (not tautology)
;       (if #f 
;           (and (> split-length 0) (>= cl-length split-length)
;                (< def-number input-size))
;           (split-clause-insert (car interm) (cadr interm) (cddr interm) 
;                                cl-length)
;           (begin (set! cl-length (min cl-length max-cl-length))
;             (vector-set! clauses-vector cl-length
;                  (cons interm (vector-ref clauses-vector cl-length))))))))

(define-macro (hash-insert! obj) `(hash-insert-key! (hash-key ,obj) ,obj))

(define-macro (hash-member? obj) `(hash-member-key? (hash-key ,obj) ,obj))

(define-macro (hashify-clause term) 
`(hashify-poslits (car ,term) (cadr ,term)))

(define-macro (max-atom atoms)
`(let (m-pos m-neg)
  (set! m-pos (a-pos-frequency (car ,atoms))) 
  (set! m-neg (a-neg-frequency (car ,atoms)))
  (if (< m-pos m-neg) 
      (max-atom-atoms (cdr ,atoms) ,atoms m-neg m-pos)
      (max-atom-atoms (cdr ,atoms) ,atoms m-pos m-neg))))

(define-macro (dyn-incr-weight clause)
`(if dynamic
     (begin
       (if (= divisor 0)
	   (set! weight (vector-ref clauses-weight-vector (c-length ,clause)))
	   (set! weight 1))
;	    (newline) (display "weight: ") (display weight)
       (increase-freq-pos (c-pos-literals ,clause))
       (increase-freq-neg (c-neg-literals ,clause))
)))

(define-macro (dyn-incr-weight1 clause)
`(if (and dynamic (> divisor 0))
     (begin (set! weight 
	       (- (vector-ref clauses-weight-vector (c-length ,clause))
		  (vector-ref clauses-weight-vector (+ 1 (c-length ,clause)))))
;	    (newline) (display "weight: ") (display weight)
	    (increase-freq-pos (c-pos-literals ,clause))
	    (increase-freq-neg (c-neg-literals ,clause))
)))

(define-macro (dyn-decr-weight clause)
`(if dynamic
     (begin
       (if (= divisor 0)
	   (set! weight 
		 (- (vector-ref clauses-weight-vector (c-length ,clause))))
	   (set! weight -1))
;	    (newline) (display "weight: ") (display weight)
       (increase-freq-pos (c-pos-literals ,clause))
       (increase-freq-neg (c-neg-literals ,clause))
)))

(define-macro (dyn-decr-weight1 clause)
`(if (and dynamic (> divisor 0))
     (begin (set! weight 
	       (- (vector-ref clauses-weight-vector (+ 1 (c-length ,clause)))
		  (vector-ref clauses-weight-vector (c-length ,clause))))
;	    (newline) (display "weight: ") (display weight)
	    (increase-freq-pos (c-pos-literals ,clause))
	    (increase-freq-neg (c-neg-literals ,clause))
)))
