%###########################################################################
%# genord.pl
%# a heuristic-based approach for generating an appropriate LPO-precedence
%# for proving termination of a given set of ordered equations
%###########################################################################
%# Elvira & Joachim Steinbach             Automated Reasoning Research Group
%#                                        Lehrstuhl VIII f"ur Informatik
%#                                        Technische Universit"at M"unchen
%# steinbac@informatik.tu-muenchen.de     D-80290 M"unchen, Germany
%########################################################################### 
%# Modified: Max Moser
%###########################################################################

:- dynamic lpo_operator_precedence/1.


%*****************************************************************************
%* genord( problemwise, +PosUnitEqs, -OperatorPrecedence )
%*****************************************************************************
%* - generate lpo operator precedence
%*****************************************************************************

genord( problemwise, ClauseList, _, NewClauseList ) :-
        listofposuniteqs( PosUnitEqs ), 
	main_genord( PosUnitEqs, OperatorSpec ),
	append( OperatorSpec, ClauseList, NewClauseList ).
	
main_genord( [], [] ) :-
	!.

main_genord( PosUnitEqs, [OperatorPrec] ) :-
	get_weight_equation_set(PosUnitEqs,[],WList),
	average_values(WList,WOut),
	insert_sort(WOut,Sorted),
	get_oplist( Sorted, OpList, NewDict ),
	OperatorPrec = rule( operator_precedence, system, [++lpo_operator_precedence(OpList)], [], NewDict ),
	assert(lpo_operator_precedence(OpList)).

%*****************************************************************************


%*****************************************************************************
%* get_weight_equation_set( +set_of_equations, +set_of_weights,
%*                          -set_of_weights )
%*****************************************************************************
%* generates a set W of weight-tuples for a set E of equations; W is of the
%* form { [f,ar,wp,wn,np,nn] | f in F } where F is the set of operators
%* induced by E, ar is the arity of f, wp (-wn) is the sum of the weights of
%* the f's occurring on the left-hand sides (right-hand sides) of the ordered
%* equations in E, np (-nn) is the number of f's occurring on the left-hand
%* sides (right-hand sides) of the ordered equations in E
%*****************************************************************************

get_weight_equation_set([],WList,WList) :-
          !.

get_weight_equation_set([E],WList,Out) :-
          get_weight_equation(E,WList,Out),
          !.

get_weight_equation_set([E|Rest],WList,Out) :-
          get_weight_equation(E,WList,EOut),
          get_weight_equation_set(Rest,EOut,Out).

%*****************************************************************************


%*****************************************************************************
%* get_weight_equation( +equation, +set_of_weights, -set_of_weights )
%*****************************************************************************
%* updates the input set of weight-tuples (2nd argument) by adding the set of
%* weight-tuples for the input equation
%*****************************************************************************
get_weight_equation(equal(Left,Right),WList,Out) :-
          get_weight_term(Left,1,1,1,WList,LOut),
          get_weight_term(Right,-1,1,1,LOut,Out),
          !.

%*****************************************************************************


%*****************************************************************************
%* get_weight_term( +term, +left_right_flag, +depth_fact, +arg_fact, 
%*                  +set_of_weights, -set_of_weights )
%*****************************************************************************
%* updates the input set of weight-tuples (5th argument) by adding the set of
%* weight-tuples for the input term;
%* the left-right-flag determines whether the current set of weight-tuples
%* must be decreased (if a right-hand side of an equation is considered) or
%* increased (if a left-hand side of an equation is considered)
%*****************************************************************************

get_weight_term(T,_,_,_,WList,WList) :-
          var(T),
          !.

get_weight_term(T,LR,Depth,Arg,WList,Out) :-
          functor(T,F,Ar),
          T =.. [F|Args],
          get_weight_op(F,Ar,LR,Depth,Arg,WList,HOut),
          RDepth is Depth/4,
          get_weight_term_list(Args,LR,RDepth,1,HOut,Out),
          !.

%*****************************************************************************


%*****************************************************************************
%* get_weight_term_list( +list_of_terms, +left_right_flag, +depth_fact,
%*                       +arg_fact, +set_of_weights, -set_of_weights )
%*****************************************************************************
%* updates the input set of weight-tuples (5th argument) by adding the set of
%* weight-tuples for the input list of terms
%*****************************************************************************

get_weight_term_list([],_,_,_,WList,WList):-
          !.

get_weight_term_list([First|Rest],LR,Depth,Arg,WList,Out) :-
          get_weight_term(First,LR,Depth,Arg,WList,FOut),
	  RArg is Arg / 4,
          get_weight_term_list(Rest,LR,Depth,RArg,FOut,Out).

%*****************************************************************************


%*****************************************************************************
%* get_weight_op( +operator, +arity, +left_right_flag, +depth_fact,
%*                       +arg_fact, +set_of_weights, -set_of_weights )
%*****************************************************************************
%* updates the input set of weight-tuples (6th argument) by adding the set of
%* weight-tuples for the input operator
%*****************************************************************************

get_weight_op(F,Ari,1,Depth,Arg,[],[[F,Ari,WP,0,1,0]]) :-
          WP is Depth * Arg.

get_weight_op(F,Ari,-1,Depth,Arg,[],[[F,Ari,0,WN,0,1]]) :-
          WN is (-1) * Depth * Arg.

get_weight_op(F,Ari,1,Depth,Arg,[[F,Ari,WP,WN,NP,NN]],[[F,Ari,WOut,WN,NP1,NN]]) :-
          WOut is WP + Depth * Arg,
          NP1 is NP + 1,
          !.

get_weight_op(F,Ari,-1,Depth,Arg,[[F,Ari,WP,WN,NP,NN]],[[F,Ari,WP,WOut,NP,NN1]]) :-
          WOut is WN + (-1) * Depth * Arg,
          NN1 is NN + 1,
          !.

get_weight_op(F,Ari,1,Depth,Arg,[[F,Ari,WP,WN,NP,NN]|Rest],[[F,Ari,WOut,WN,NP1,NN]|Rest]) :-
          WOut is WP + Depth * Arg,
          NP1 is NP + 1,
          !.

get_weight_op(F,Ari,-1,Depth,Arg,[[F,Ari,WP,WN,NP,NN]|Rest],[[F,Ari,WP,WOut,NP,NN1]|Rest]) :-
          WOut is WN + (-1) * Depth * Arg,
          NN1 is NN + 1,
          !.

get_weight_op(F,Ari,LR,Depth,Arg,[B|Rest],[B|ROut]) :-
          get_weight_op(F,Ari,LR,Depth,Arg,Rest,ROut),
          !.

get_weight_op(F,Ari,1,Depth,Arg,WList,[[F,Ari,WOut,0,1,0]|WList]) :-
          WOut is Depth * Arg.

get_weight_op(F,Ari,-1,Depth,Arg,WList,[[F,Ari,0,WOut,0,1]|WList]) :-
          WOut is (-1) * Depth * Arg.

%*****************************************************************************


%*****************************************************************************
%* average_values( +set_of_weights, -set_of_weights )
%*****************************************************************************
%* computes the final weight-tuple for each operator in the following way: let
%* [f,ar,wp,wn,np,nn] be the weight-tuple for the operator f; then
%* [f,ar,(wp/np)+(wn/nn)] is the new weight-tuple
%*****************************************************************************

average_values([],[]) :-
          !.

average_values([First|Rest],WOut) :-
          First = [F,Ar,WP,WN,NP,NN],
          div_0(WP,NP,RP),
          div_0(WN,NN,RN),
          V is RP + RN,
          average_values(Rest,ROut),
          append([[F,Ar,V]],ROut,WOut).

%*****************************************************************************


%*****************************************************************************
%* div_0( +int, +int, -rat )
%*****************************************************************************
%* divides the first argument by the second one by returning 0 if the second
%* argument is 0
%*****************************************************************************

div_0(_,0,0) :-
          !.

div_0(X,Y,Z) :-
          Z is X/Y.

%*****************************************************************************


%*****************************************************************************
%* insert_sort( +set_of_weight, -set_of_weight )
%*****************************************************************************
%* sorts the input set { [f,ar,w] | f in F } of weight-tuples wrt. w in
%* descending order
%*****************************************************************************

insert_sort([],[]).

insert_sort([X|Tail],Sorted) :-
          insert_sort(Tail,SortedTail),
          insert(X,SortedTail,Sorted).

%*****************************************************************************


%*****************************************************************************
%* insert( +item, +set_of_weight, -set_of_weight )
%*****************************************************************************
%* inserts the head X into the sorted list [Y|Sorted]
%*****************************************************************************

insert(X,[Y|Sorted],[Y|Sorted1]) :-
          X = [_,_,WX],
          Y = [_,_,WY],
          WX < WY,
          !,
          insert(X,Sorted,Sorted1).

insert(X,Sorted,[X|Sorted]).

%*****************************************************************************


%*****************************************************************************
%* get_oplist( +SetOfWeights, -OpList, -NewDict )
%*****************************************************************************
%* returns the precedence based on the (sorted) set of weight-tuples
%*****************************************************************************

get_oplist( [], [], [] ).

get_oplist( [[F,Ar,_]|Rest], [AnonTerm|RestOpList], AllNewDicts ) :-
	get_n_vars_and_anon_dict( Ar, ListOfNewVars, NewDict ),
	AnonTerm =.. [F|ListOfNewVars],
	get_oplist( Rest, RestOpList, RestNewDicts ),
	append( NewDict, RestNewDicts, AllNewDicts ).

%*****************************************************************************


%*****************************************************************************
%* get_n_vars_and_anon_dict( +Ar, -ListOfNewVars, -NewDict )
%*****************************************************************************
%* returns a list of Ar new variables and generate dictionary
%*****************************************************************************

get_n_vars_and_anon_dict( 0, [], [] ).

get_n_vars_and_anon_dict( N, [NewVar|RestUnderscores], [[NewVar|'_']|RestDict] ) :-
	getnewvar( NewVar ),

	N>0,
	N1 is N-1,
	get_n_vars_and_anon_dict( N1, RestUnderscores, RestDict ).


%*****************************************************************************
%* - check whether ordering addition really needs to be performed
%*****************************************************************************

test_switches_genord :-
		% if formula contains no equality
	predaritylist( PredArityList ),
	\+ member( equal/_, PredArityList ),
	!,
		% then remove all ordering additions
	retract_all(cmd_arg(genord,_)).

test_switches_genord.

%*****************************************************************************
