%###########################################################################
%# proc_loop.pl
%# - main processing loop
%###########################################################################
%# predicates to be used from outside:
%#	- process_problem/2
%# dependent on
%#	- mod_*.pl
%#	- mylib.pl
%###########################################################################
%#        Max Moser                         Automated Reasoning Research Group
%# Email: moser@informatik.tu-muenchen.de     Lehrstuhl VIII f"ur Informatik
%# Tel.:  +49-89/521096			     Technische Universit"at M"unchen
%# Fax.:  +49-89/526502			         D-80290 M"unchen, Germany
%# http://wwwjessen.informatik.tu-muenchen.de/~moser
%###########################################################################



%*****************************************************************************
%* process_problem( +ClauseList, -NewClauseList )
%*****************************************************************************
%* - process the whole clause list ClauseList 
%*   wrt to the modifications given 
%* - return result in NewClauseList
%* - clausewise(<mod>) means that each clause is processed by itself
%* - problemwise(<mod>) means that modification is to be called with the
%*		whole problem
%*****************************************************************************

process_problem( ClauseList, FullClauseList ) :-
	filterseq( FilterSeqList ),
	domod( ClauseList, FilterSeqList, FullClauseList ).

%*****************************************************************************


%*****************************************************************************
%* domod( +ClauseList, +Mods, -NewClauseList )
%*****************************************************************************
%* - for each modification
%*	. check whether modification needs to be done
%*	. actually do the modification to all the current clauses
%*	. continue with the resulting clauses
%*****************************************************************************

domod( ClauseList, [], ClauseList ).
		% no more modifications left. we are done.

domod( ClauseList, [clausewise(Mod)|RestMods], NewClauseList ) :-
		% if modification for each line needs to be done
	cmd_arg( Mod, SwitchList ),
	!,
	on_verbose_report( trying(Mod,clausewise), "-" ),
		% call current modification. result is ErgClauseList
	domodify( ClauseList, Mod, SwitchList, ErgClauseList ),
	on_verbose_report( '-done', "\n" ),
		% apply remaining modifications to ErgClauseList
	domod( ErgClauseList, RestMods, NewClauseList ).

domod( ClauseList, [problemwise(Mod)|RestMods], NewClauseList ) :-
		% if modification to whole file needs to be done
	cmd_arg( Mod, SwitchList ),
	!,
	on_verbose_report( trying(Mod,problemwise), "-" ),
		% construct call for Mod and SwitchList and current clause
	Call =.. [Mod,problemwise,ClauseList,SwitchList,ErgClauseList],
		% execute call
	call( Call ),
		% apply remaining modifications to ErgClauseList
	on_verbose_report( done, "\n" ),
	domod( ErgClauseList, RestMods, NewClauseList ).

domod( ClauseList, [_|RestMods], NewClauseList ) :-
		% modification does not need to be done
	domod( ClauseList, RestMods, NewClauseList ).

%*****************************************************************************


%*****************************************************************************
%* domodify( +ClauseList, +Mod, +SwitchList, -NewClauseList )
%*****************************************************************************
%* - for each clause in clause list:
%*   construct a call for respective modification and execute call
%* - collect all clauses
%*****************************************************************************

domodify( [], _, _, [] ).
		% no clauses left to modify. we are done

domodify( [Clause|RestClauses], Mod, SwitchList, NewClauseList ) :-
		% construct call for Mod and SwitchList and current clause
	Call =.. [Mod,clausewise,Clause,SwitchList,ErgClauseList],
		% execute call
	call( Call ),
		% continue with remaining clauses
	on_verbose_report( '.', "" ),
	domodify( RestClauses, Mod, SwitchList, NewClauseListRest ),
		% collect all the results
	append( ErgClauseList, NewClauseListRest, NewClauseList ).

%*****************************************************************************
