%###########################################################################
%# mylib.pl
%# - various useful utilities for manipulating clauses
%###########################################################################
%# predicates to be used from outside:
%#	- delete_ordered_n( ?SubList, +FullList, +Number, ?Rest )
%#	- get_list_of_n( +N, +WhatEver, -ListOfNWhatEver )
%#	- last_elem( +List, ?LastElem )
%#	- int2s( +IntVal, -SVal )
%#	- s2int( +SVal, -IntVal )
%#	- inst_dict( ?Dict )
%#	- is_bip( +Lit )
%#	- make_dict( +EclipseDict, -MyDict )
%#	- litlist_to_posneglist( +LitList, -PosList, -NegList )
%#	- posneglist_to_litlist( +PosList, +NegList, -LitList )
%#	- conjlits2litlist( +ConjLits, -LitList )
%#	- disjlits2litlist( +ConjLits, -LitList )
%#	- addinfo2newclauselist( +Name, +Type, +LitListList, +Cons, +Dict, -ClauseList)
%#	- addinfo2newrulelist( +Name, +Type, +LitListList, +Cons, +Dict, -RuleList)
%#	- getnewvar( -NewVar )
%#	- add_index_to_atom( +Atom, +Index, -AtomWithIndex )
%#	- fetch_next_token( +File, -Token )
%#	- gen_tmp_filename( -TmpFileName )
%#	- cp_file_from_to( +InStream, +OutStream )
%#	- add_prefix_until_unique_in( +OldAtom, +Prefix, +SymbArityList, -NewAtom )
%#	- prependtoall( +Term, +ListOfLists, -Result )
%#	- gen_corr_uc_symbol( +LCAtom, +Number, -UCAtom )
%#	- only_negative_literals( +LitList )
%#	- only_positive_literals( +LitList )
%#	- gen_new_variant( +Something, -NewVariantOfSomething )
%#	- gen_new_variants_list( +SomethingList, -NewVariantOfSomethingList )
%#	- sprintf( -OutString, +FormatString, +ParList )
%#	- saveit( +MainCall, +NameOfSaveState )
%###########################################################################
%#        Max Moser                         Automated Reasoning Research Group
%# Email: moser@informatik.tu-muenchen.de     Lehrstuhl VIII f"ur Informatik
%# Tel.:  +49-89/521096			     Technische Universit"at M"unchen
%# Fax.:  +49-89/526502			         D-80290 M"unchen, Germany
%# http://wwwjessen.informatik.tu-muenchen.de/~moser
%###########################################################################



%*****************************************************************************
%* delete_ordered_n( ?SubList, +FullList, +Number, ?Rest )
%*****************************************************************************
%* - delete list SubList of Number elements from FullList giving Rest
%* - the ordering of the elements in SubList is the same as in FullList
%*****************************************************************************

delete_ordered_n( [], RestAll, 0, RestAll ) :-
		% 0 elements to delete so we are done
	!.

delete_ordered_n( [Del|RestDel], [Del|RestAll], N, Rest ) :-
		% are still elements to be deleted?
	N > 0,
	N1 is N-1,
		% continue to delete N-1 elements from RestAll
	delete_ordered_n( RestDel, RestAll, N1, Rest ).

delete_ordered_n( RestDel, [Keep|RestAll], N, [Keep|Rest] ) :-
		% continue to delete N elements from RestAll
	delete_ordered_n( RestDel, RestAll, N, Rest ).

%*****************************************************************************


%*****************************************************************************
%* get_list_of_n( +N, +WhatEver, -ListOfNWhatEver )
%*****************************************************************************
%* - produce a list of N terms WhatEver
%*****************************************************************************

get_list_of_n( 0, _, [] ).

get_list_of_n( N, WhatEver, [WhatEver|RestListOfNWhatEver] ) :-
	N > 0,
	N1 is N-1,
	get_list_of_n( N1, WhatEver, RestListOfNWhatEver ).

%*****************************************************************************


%*****************************************************************************
%* last_elem( +List, ?LastElem )
%*****************************************************************************
%* - find last element in list
%*****************************************************************************

last_elem( [LastElem], LastElem ) :-
	!.

last_elem( [_|Rest], LastElem ) :-
	last_elem( Rest, LastElem ).

%*****************************************************************************


%*****************************************************************************
%* int2s( +IntVal, -SVal )
%* s2int( +SVal, -IntVal )
%*****************************************************************************
%* - convert integer to s^n(0) and vice versa
%*****************************************************************************

int2s( 0, 0 ) :-
	!.
int2s( N, s(Sn) ) :-
	N > 0,
	N1 is N -1,
	int2s( N1, Sn ).

s2int( 0, 0 ) :-
	!.
s2int( s(Sn), N ) :-
	s2int( Sn, N1 ),
	N is N1+1.

%*****************************************************************************


%*****************************************************************************
%* inst_dict( ?Dict )
%*****************************************************************************
%* - instantiate all variables contained in Dict 
%*****************************************************************************

inst_dict( [] ).
inst_dict( [[Term|Term]|Rest] ) :- 
	!,
	inst_dict( Rest ).
inst_dict( [Term=Term|Rest] ) :- 
	!,
	inst_dict( Rest ).
inst_dict( [_|Rest] ) :- 
	inst_dict( Rest ).

%*****************************************************************************


%*****************************************************************************
%* is_bip( +Lit )
%*****************************************************************************
%* - test whether predicate symbol of literal starts with '$' or
%    contains a blank (e.g. 'X is 3')
%*****************************************************************************

is_bip( Lit ) :-
	Lit =.. [PSymb|_],
 	name( PSymb, [36|_] ),
	!.

is_bip( Lit ) :-
	atomic( Lit ),
	name( Lit, CharList ),
	member( 32, CharList ).

%*****************************************************************************


%*****************************************************************************
%* make_dict( +EclipseDict, -MyDict )
%*****************************************************************************
%* - converty representation of dictionary
%*****************************************************************************

make_dict([],[]).

make_dict([[Name|Variable]|Rest],[Name=Variable|RestConverted]):-
    make_dict(Rest,RestConverted).

%*****************************************************************************


%*****************************************************************************
%* litlist_to_posneglist( +LitList, -PosList, -NegList )
%*****************************************************************************
%* - split list of ++/-- literals into list of pos and list of neg literals
%*****************************************************************************

litlist_to_posneglist( [], [], [] ).

litlist_to_posneglist( [++Lit|Rest], [Lit|PosRest], NegRest ) :-
	litlist_to_posneglist( Rest, PosRest, NegRest ).

litlist_to_posneglist( [--Lit|Rest], PosRest, [Lit|NegRest] ) :-
	litlist_to_posneglist( Rest, PosRest, NegRest ).

%*****************************************************************************


%*****************************************************************************
%* posneglist_to_litlist( +PosList, +NegList, -LitList )
%*****************************************************************************
%* - join list of pos and list of neg literals to list of ++/-- literals
%*****************************************************************************

posneglist_to_litlist( [~Lit|PosRest], NegRest, [--Lit|Rest] ) :-
	!,
	posneglist_to_litlist( PosRest, NegRest, Rest ).

posneglist_to_litlist( [Lit|PosRest], NegRest, [++Lit|Rest] ) :-
	posneglist_to_litlist( PosRest, NegRest, Rest ).

posneglist_to_litlist( [], [~Lit|NegRest], [++Lit|Rest] ) :-
	!,
	posneglist_to_litlist( [], NegRest, Rest ).

posneglist_to_litlist( [], [Lit|NegRest], [--Lit|Rest] ) :-
	posneglist_to_litlist( [], NegRest, Rest ).

posneglist_to_litlist( [], [], [] ).

%*****************************************************************************


%*****************************************************************************
%* conjlits2litlist( +ConjLits, -LitList )
%*****************************************************************************
%* - convert (Lit,...,Lit) to Prolog list [Lit,...,Lit]
%*****************************************************************************

conjlits2litlist( (Lit,RestLits), [Lit|RestLitList] ) :-
	!,
	conjlits2litlist( RestLits, RestLitList ).

conjlits2litlist( Lit, [Lit] ).

%*****************************************************************************


%*****************************************************************************
%* disjlits2litlist( +ConjLits, -LitList )
%*****************************************************************************
%* - convert (Lit;...;Lit) to Prolog list [Lit,...,Lit]
%*****************************************************************************

disjlits2litlist( (Lit;RestLits), [Lit|RestLitList] ) :-
	!,
	disjlits2litlist( RestLits, RestLitList ).

disjlits2litlist( Lit, [Lit] ).

%*****************************************************************************


%*****************************************************************************
%* addinfo2newclauselist( +Name, +Type, +LitListList, +Cons, +Dict, -ClauseList)
%* addinfo2newrulelist( +Name, +Type, +LitListList, +Cons, +Dict, -RuleList)
%*****************************************************************************
%* - generate list of clauses with corresponding information
%*****************************************************************************

addinfo2newclauselist( _, _, [], _, _, [] ).

addinfo2newclauselist( Name, Type, [LitList|RestLitListList], Cons, Dict, 
							NewClauseList) :-
	NewClause = clause( Name, Type, LitList, Cons, Dict ),
	addinfo2newclauselist( Name, Type, RestLitListList, Cons, Dict, 
							NewClauseListRest),
	NewClauseList = [NewClause|NewClauseListRest].

%*****************************************************************************

addinfo2newrulelist( _, _, [], _, _, [] ).

addinfo2newrulelist( Name, Type, [LitList|RestLitListList], Cons, Dict, 
							NewRuleList) :-
	NewRule = rule( Name, Type, LitList, Cons, Dict ),
	addinfo2newrulelist( Name, Type, RestLitListList, Cons, Dict, 
							NewRuleListRest),
	NewRuleList = [NewRule|NewRuleListRest].
	
%*****************************************************************************


%*****************************************************************************
%* getnewvar( -NewVar )
%*****************************************************************************
%* - return new Prolog Variable
%*****************************************************************************

getnewvar( _ ).

%*****************************************************************************


%*****************************************************************************
%* add_index_to_atom( +Atom, +Index, -AtomWithIndex )
%*****************************************************************************
%* - e.g. ('clause',5,'clause__5__') or ('clause',-3,'clause__minus_3__')
%*****************************************************************************

add_index_to_atom( Atom, Index, AtomWithIndex ) :-
		% in case of negative numbers ...
	Index < 0,
	!,
		% generate character list for given atom
	name( Atom, AtomChars ),
		% generate character list for '__minus_'
	name( '__minus_', SepChars ),
		% generate character list for positive index
	PosIndex is -Index,
	name( PosIndex, IndexChars ),
		% generate character list for '__'
	name( '__', EndChars ),
		% append character lists for atom, '__minus_' and pos index
	append( IndexChars, EndChars, TmpChars1 ),
	append( SepChars, TmpChars1, TmpChars2 ),
	append( AtomChars, TmpChars2, TmpChars3 ),
		% and finally construct new atom
	name( AtomWithIndex, TmpChars3 ).

add_index_to_atom( Atom, Index, AtomWithIndex ) :-
		% in case of positive numbers ...
		% generate character list for given atom
	name( Atom, AtomChars ),
		% generate character list for '__'
	name( '__', SepChars ),
		% generate character list for index
	name( Index, IndexChars ),
		% append character lists for atom, '__' and index
	append( IndexChars, SepChars, TmpChars1 ),
	append( SepChars, TmpChars1, TmpChars2 ),
	append( AtomChars, TmpChars2, TmpChars3 ),
		% and finally construct new atom
	name( AtomWithIndex, TmpChars3 ).

%*****************************************************************************


%*****************************************************************************
%* fetch_next_token( +File, -Token )
%*****************************************************************************
%* - read next token from input file
%* - reset stream pointer to initial position
%* - NOTE: This predicate is highly dependent on Eclipse-Prolog!!!!!
%*****************************************************************************

fetch_next_token( File, Token ) :-
		% remember current position in input file
	at( File, CurrentPositionInFile ),
		% get next token
	read_token( File, Token, _ ),
		% reset position in input file
	seek( File, CurrentPositionInFile ).

%*****************************************************************************


%*****************************************************************************
%* gen_tmp_filename( -TmpFileName )
%*****************************************************************************
%* - generate temporary file name using pid of current process
%* - use counter tmpfiles/1 for different temporary files in same programm
%*****************************************************************************

:- dynamic tmpfiles/1.
:- assert( tmpfiles( 0 ) ).
	% up to now, no temporary files have been generated 

gen_tmp_filename( TmpFileName ) :-
		% generate initial name prefix
	name( '/tmp/stexposu_', IniNameASCII ),
		% get process id
	get_flag( pid, CurrentProcessId ),
	name( CurrentProcessId, CurrentProcessIdASCII ),
		% get number of temporary files
	retract(tmpfiles( TmpFileNum )),
	NewTmpFileNum is TmpFileNum+1,
	assert(tmpfiles(NewTmpFileNum)),
	name( TmpFileNum, TmpFileNumASCII ),
		% get ASCII for '_'
	name( '_', UnderscoreASCII ),
		% append all ASCII lists and generate TmpFileName
	append( UnderscoreASCII, TmpFileNumASCII, Tmp1ASCII ),
	append( CurrentProcessIdASCII, Tmp1ASCII, Tmp2ASCII ),
	append( IniNameASCII, Tmp2ASCII, TmpFileNameASCII ),
	name( TmpFileName, TmpFileNameASCII ).

%*****************************************************************************


%*****************************************************************************
%* cp_file_from_to( +InStream, +OutStream )
%*****************************************************************************
%* - copy all characters from InStream to OutStream
%* - streams are assumed to be open files
%*****************************************************************************

cp_file_from_to( InStream, OutStream ) :-
		% get character from InStream
	get( InStream, Char ),
		% eventually write character to OutStream
	cp_file_from_to_char( Char, InStream, OutStream ).

cp_file_from_to_char( -1, _, _ ) :-
		% if -1 has been read then we are at end of file
	!.

cp_file_from_to_char( Char, InStream, OutStream ) :-
		% not at end of file, write character and continue
	put( OutStream, Char ),
	cp_file_from_to( InStream, OutStream ).

%*****************************************************************************


%*****************************************************************************
%* add_prefix_until_unique_in( +OldAtom, +Prefix, +SymbArityList, -NewAtom )
%*****************************************************************************
%* - add Prefix to OldAtom until NewAtom does not uccur in SymbArityList
%* - SymbArityList is a list of entries Symbol/Arity
%*****************************************************************************

add_prefix_until_unique_in( OldAtom, Prefix, SymbArityList, NewAtom ) :-
		% if OldAtom a known symbol ...
	member( OldAtom/_, SymbArityList ),
	!,
		% then add Prefix to OldAtom
	name( Prefix, PrefixASCII ),
	name( OldAtom, OldAtomASCII ),
	append( PrefixASCII, OldAtomASCII, TmpAtomASCII ),
	name( TmpAtom, TmpAtomASCII ),
		% try new atom
	add_prefix_until_unique_in( TmpAtom, Prefix, SymbArityList, NewAtom ).

add_prefix_until_unique_in( OldAtom, _, _, OldAtom ).
		% else OldAtom is a new symbol 

%*****************************************************************************


%*****************************************************************************
%* prependtoall( +Term, +ListOfLists, -Result )
%*****************************************************************************
%* - add Term as new head of all lists in ListOfLists
%*****************************************************************************

prependtoall( _, [], [] ).

prependtoall( Term, [List|RestLists], [[Term|List]|PrepRest] ) :-
        prependtoall( Term, RestLists, PrepRest ).

%*****************************************************************************


%*****************************************************************************
%* gen_corr_uc_symbol( +LCAtom, +Number, -UCAtom )
%*****************************************************************************
%* - convert all lower case chars in LCAtom to upper case
%* - if desired (Number>0) append also the given number
%*****************************************************************************

% from STEXPOSU_NEW (incomplete) Joachim 15.07.98
gen_corr_uc_symbol( LCAtom, 0, UCAtom ) :-
        cmd_arg( special, _ ),
        var( LCAtom ),
        !,
        getnewvar(UCAtom).

gen_corr_uc_symbol( LCAtom, 0, UCAtom ) :-
	integer( LCAtom ),
	!,
	name( LCAtom, LCAtomAsciis ),
	asciis_lc2uc( LCAtomAsciis, UCAtomAsciis ),
	name( 'Int', H ),
	append( H, UCAtomAsciis, AllUCAtomAsciis ),
	name( UCAtom, AllUCAtomAsciis ).

gen_corr_uc_symbol( LCAtom, 0, UCAtom ) :-
	!,
	name( LCAtom, LCAtomAsciis ),
	asciis_lc2uc( LCAtomAsciis, UCAtomAsciis ),
	name( UCAtom, UCAtomAsciis ).

gen_corr_uc_symbol( LCAtom, Number, UCAtom ) :-
	integer( LCAtom ),
	!,
	name( LCAtom, LCAtomAsciis ),
	asciis_lc2uc( LCAtomAsciis, UCAtomAsciis ),
	name( 'Int', H ),
	name( Number, NumberInAscii ),
	name( '_', [US_Ascii] ),
	append( UCAtomAsciis, [US_Ascii|NumberInAscii], NumberAtomAsciis ),
	append( H, NumberAtomAsciis, FullNumberAtomAsciis ),
	name( UCAtom, FullNumberAtomAsciis ).

gen_corr_uc_symbol( LCAtom, Number, UCAtom ) :-
	name( LCAtom, LCAtomAsciis ),
	asciis_lc2uc( LCAtomAsciis, UCAtomAsciis ),
	name( Number, NumberInAscii ),
	name( '_', [US_Ascii] ),
	append( UCAtomAsciis, [US_Ascii|NumberInAscii], NumberAtomAsciis ),
	name( UCAtom, NumberAtomAsciis ).

asciis_lc2uc( [], [] ).
asciis_lc2uc( [LCAscii|RestAsciis], [UCAscii|RestUCAsciis] ) :-
	97 =< LCAscii, LCAscii =< 122,
	!,
	UCAscii is LCAscii - 32,
	asciis_lc2uc( RestAsciis, RestUCAsciis ).
asciis_lc2uc( [Ascii|RestAsciis], [Ascii|RestUCAsciis] ) :-
	asciis_lc2uc( RestAsciis, RestUCAsciis ).

%*****************************************************************************


%*****************************************************************************
%* only_negative_literals( +LitList )
%*****************************************************************************
%* - test whether LitList contains only negative (--) literals
%*****************************************************************************

only_negative_literals( [] ).

only_negative_literals( [--_|RestLitList] ) :-
        only_negative_literals( RestLitList ).

%*****************************************************************************


%*****************************************************************************
%* only_positive_literals( +LitList )
%*****************************************************************************
%* - test whether LitList contains only positive (--) literals
%*****************************************************************************

only_positive_literals( [] ).

only_positive_literals( [++_|RestLitList] ) :-
        only_positive_literals( RestLitList ).

%*****************************************************************************


%*****************************************************************************
%* gen_new_variant( +Something, -NewVariantOfSomething )
%* gen_new_variants_list( +SomethingList, -NewVariantOfSomethingList )
%*****************************************************************************
%* - generate new variants of items
%*****************************************************************************

gen_new_variants_list( [], [] ).
		% well, no items left so we are done

gen_new_variants_list( [Item|RestItems], [NewVarItem|NewVarRestItems] ) :-
	gen_new_variant( Item, NewVarItem ),
	gen_new_variants_list( RestItems, NewVarRestItems ).


:- dynamic temporary_buffer_for_variants/1.


gen_new_variant( Item, NewVarItem ) :-
	retract_all(temporary_buffer_for_variants(_)),
	assert(temporary_buffer_for_variants(Item)),
	temporary_buffer_for_variants(NewVarItem).

%*****************************************************************************


%*****************************************************************************
%* sprintf( -OutString, +FormatString, +ParList )
%*****************************************************************************
%* - formatted printing to OutString
%*****************************************************************************

sprintf( OutString, FormatString, ParList ) :-
	open( "", string, s ),
	printf( s, FormatString, ParList ),
	current_stream( OutString, string, s ),
	close( s ).

%*****************************************************************************


%*****************************************************************************
%* saveit( +MainCall, +NameOfSaveState )
%*****************************************************************************
%* - generate save state
%*****************************************************************************

saveit( MainCall, NameOfSaveState ) :-
	save(NameOfSaveState),
	call(MainCall).

saveit( _, _ ) :-
	exit(1).

%*****************************************************************************
